import boto3
import yaml
import requests
from datetime import datetime
import simplejson as json


def current_user(profile_name):
    sts = boto3.client('sts')
    return sts.get_caller_identity()['Arn']


def get_parameters(stack_file, stack_name, secret, stack_region=None,
                   stack_account=None, action=None,
                   profile_name=None, slack_webhook_url=None):

    stacks = None
    try:
        with open(stack_file, 'r') as stack_stream:
            stacks = yaml.safe_load(stack_stream)
    except yaml.YAMLError as exc:
        print(exc)
        alert(stack_name, exc, stack_region, stack_account, action,
              profile_name, slack_webhook_url)
        exit(1)

    template = None
    try:
        with open(stacks[stack_name]['Template'], 'r') as template_stream:
            template = json.load(template_stream)
    except Exception as e:
        try:
            with open(stacks[stack_name]['Template'], 'r') as template_stream:
                template = yaml.safe_load(template_stream)
        except yaml.YAMLError as exc:
            print(exc)
            exit(1)

    # Get list of required parameters for stack
    required_params = []
    try:
        required_params = [k for k, v in template.get('Parameters', {}).items()]
    except Exception as exc:
        print(exc)
        alert(stack_name, "Unable to get list of required parameters. Exception = {}".format(str(exc)),
              stack_region, stack_account, action,
              profile_name, slack_webhook_url)
        exit(1)

    # Fill in output array with Parameters from stack file
    output_array = []
    try:
        if "Parameters" in template:
            for key in template['Parameters']:
                if key in stacks[stack_name]['Parameters']:
                    output_array.append({
                        "ParameterKey": key,
                        "ParameterValue": stacks[stack_name]['Parameters'][key]
                    })
    except Exception as exc:
        print(exc)
        alert(stack_name,
              "Unable to get parameter key/value definition from stack file. Exception = {}".format(str(exc)),
              stack_region, stack_account, action,
              profile_name, slack_webhook_url)
        exit(1)

    # Get list of secret arguments
    array_parameters_from_arguments = []
    try:
        no_echo_overrides = dict([pair.split('=') for pair in secret or []])
        for key in no_echo_overrides:
            array_parameters_from_arguments.append({
                    "ParameterKey": key,
                    "ParameterValue": no_echo_overrides[key]
                })
    except Exception as exc:
        print(exc)
        alert(stack_name,
              "Unable to get read of secret parameters from arguments. Exception = {}".format(str(exc)),
              stack_region, stack_account, action,
              profile_name, slack_webhook_url)
        exit(1)

    # If key with same name as secret is present in stack_file, overwrite it in output_array
    # If key with same name as secret is NOT present in stack_file, then just add it to output_array
    try:
        for i in range(len(array_parameters_from_arguments)):
            found = False
            for j in range(len(output_array)):
                if array_parameters_from_arguments[i]["ParameterKey"] == output_array[j]["ParameterKey"]:
                    output_array[j]["ParameterValue"] = array_parameters_from_arguments[i]["ParameterValue"]
                    found = True
                    break
            if not found:
                output_array.append({
                    "ParameterKey": array_parameters_from_arguments[i]["ParameterKey"],
                    "ParameterValue": array_parameters_from_arguments[i]["ParameterValue"]
                })
    except Exception as exc:
        print(exc)
        alert(stack_name,
              "Unable to get concatenate secret and stack file parameter definition. Exception = {}".format(str(exc)),
              stack_region, stack_account, action,
              profile_name, slack_webhook_url)
        exit(1)

    # If any template parameters are still required but not defined, then push them with "UsePreviousValue"
    try:
        for param in required_params:
            found = False
            for i in range(len(output_array)):
                if output_array[i]["ParameterKey"] == param:
                    found = True
                    break
            if not found:
                output_array.append({
                    "ParameterKey": param,
                    "UsePreviousValue": True
                })
    except Exception as exc:
        print(exc)
        alert(stack_name,
              "Unable to set non-defined parameters with UsePreviousValue logic. Exception = {}".format(str(exc)),
              stack_region, stack_account, action,
              profile_name, slack_webhook_url)
        exit(1)

    return output_array


def get_tags(stack_file, stack_name, stack_region=None,
             stack_account=None, action=None,
             profile_name=None, slack_webhook_url=None):

    stacks = None
    try:
        with open(stack_file, 'r') as stack_stream:
            stacks = yaml.safe_load(stack_stream)
    except yaml.YAMLError as exc:
        print(exc)
        alert(stack_name, exc, stack_region, stack_account, action,
              profile_name, slack_webhook_url)
        exit(1)

    tags = []
    try:
        tags = stacks[stack_name]['Tags']
    except Exception as exc:
        print(exc)
        alert(stack_name,
              "Unable to set tags based on stack file. Exception = {}".format(str(exc)),
              stack_region, stack_account, action,
              profile_name, slack_webhook_url)
        exit(1)

    return [{'Key': k, 'Value': v} for k, v in tags.items()]


def upload(stack_file, stack_name, s3_bucket, s3_prefix,
           stack_region=None, stack_account=None, action=None,
           profile_name=None, slack_webhook_url=None):

    stacks = None
    try:
        with open(stack_file, 'r') as stack_stream:
            stacks = yaml.safe_load(stack_stream)
    except yaml.YAMLError as exc:
        print(exc)
        alert(stack_name, exc, stack_region, stack_account, action,
              profile_name, slack_webhook_url)
        exit(1)

    template_string = None
    try:
        with open(stacks[stack_name]['Template'], 'r') as template_stream:
            template_string = json.load(template_stream)
    except Exception as e:
        try:
            with open(stacks[stack_name]['Template'], 'r') as template_stream:
                template_string = yaml.safe_load(template_stream)
        except yaml.YAMLError as exc:
            print(exc)
            exit(1)

    output_object = {
        "type": "TemplateBody",
        "value": json.dumps(template_string)
    }

    template_key = None
    if s3_bucket is not None and stack_region is not None:
        timestamp_key = datetime.utcnow().isoformat().replace(":", "-").replace(".", "-")
        template_key = '{}/{}/{}'.format(stack_name, timestamp_key, stacks[stack_name]['Template'])
        if s3_prefix is not None:
            while str(s3_prefix).endswith("/"):
                s3_prefix = str(s3_prefix)[:-1]
            template_key = '{}/{}/{}/{}'.format(s3_prefix, stack_name, timestamp_key,
                                                stacks[stack_name]['Template'])
        try:
            s3 = boto3.resource('s3', region_name=stack_region)
            s3.Bucket(s3_bucket).upload_file(
                stacks[stack_name]['Template'],
                template_key
            )
        except Exception as exc:
            print(exc)
            alert(stack_name,
                  "Unable to push template to S3 bucket {} and S3 prefix {}. Exception = {}".format(
                      s3_bucket, s3_prefix, str(exc)
                  ),
                  stack_region, stack_account, action,
                  profile_name, slack_webhook_url)
            exit(1)

        output_object = {
            "type": "TemplateURL",
            "value": "https://{}.s3.amazonaws.com/{}".format(s3_bucket, template_key)
        }

    return output_object


def result(stack_file, stack_name, secret, s3_bucket, s3_prefix,
           stack_region=None, stack_account=None, action=None,
           profile_name=None, slack_webhook_url=None):

    params = get_parameters(stack_file, stack_name, secret, stack_region,
                            stack_account, action, profile_name, slack_webhook_url)

    tags = get_tags(stack_file, stack_name, stack_region,
                    stack_account, action, profile_name, slack_webhook_url)

    template_url = upload(stack_file, stack_name, s3_bucket, s3_prefix, stack_region,
                    stack_account, action, profile_name, slack_webhook_url)

    return params, tags, template_url


def alert(stack_name, error=None, stack_region=None, stack_account=None, action=None,
          profile_name=None, slack_webhook_url=None):

    if slack_webhook_url is not None:
        iam_user = current_user(profile_name)
        action = str(action).upper()

        if stack_region is not None and stack_account is not None:
            tpl = None
            payload_text = '*{}* - *{}*'.format(stack_name, action)
            if error is not None:
                tpl = 'Stack *{}* process failed with error *{}* for stackId ' \
                      '<https://console.aws.amazon.com/cloudformation/home?region={}#/stacks/stackinfo?stackId={}>' \
                      ' in region `{}` for account Id `{}`.'.format(
                        action, error, stack_region,
                        stack_name, stack_region, stack_account)

                payload_text = '*{}* - *{}* - *FAILED*'.format(stack_name, str(action).upper())
            else:
                tpl = 'IAM user `{}` is doing operation *{}* for stack ' \
                      '<https://console.aws.amazon.com/cloudformation/home?region={}#/stacks/stackinfo?stackId={}>' \
                      ' in region `{}` for account Id `{}`.'.format(
                        iam_user, action, stack_region,
                        stack_name, stack_region, stack_account)

            payload = {
                'text': payload_text,
                'attachments': [
                    {
                        'fallback': tpl,
                        'fields': [
                            {
                                "title": "Stack URL",
                                "value": 'https://console.aws.amazon.com/cloudformation/home?region={}#'
                                         '/stacks/stackinfo?stackId={}'.format(stack_region, stack_name),
                                "short": False
                            },
                            {
                                "title": "AWS IAM Author",
                                "value": iam_user,
                                "short": False
                            },
                            {
                                "title": "AWS Region",
                                "value": stack_region,
                                "short": False
                            },
                            {
                                "title": "AWS Account Id",
                                "value": stack_account,
                                "short": False
                            },
                            {
                                "title": "Operation",
                                "value": str(action).upper(),
                                "short": False
                            }
                        ]
                    }

                ]
            }
            if error is not None:
                payload['attachments'][0]['fields'].append({
                    "title": "Error",
                    "value": error,
                    "short": False
                })
            try:
                requests.post(slack_webhook_url, json=payload)
            except Exception as e:
                print(e)
                return False
        else:
            tpl = None
            payload_text = '*{}* - *{}*'.format(stack_name, action)
            if error is not None:
                tpl = 'Stack operation *{}* process failed with error *{}* for stackId `{}`.'.format(
                    action, error, stack_name)
                payload_text = '*{}* - *{}* - *FAILED*'.format(stack_name, action)
            else:
                tpl = 'IAM user `{}` is doing operation *{}* for stackId `{}`.'.format(
                    iam_user, action, stack_name)

            payload = {
                'text': payload_text,
                'attachments': [
                    {
                        'fallback': tpl,
                        'fields': [
                            {
                                "title": "AWS IAM Author",
                                "value": iam_user,
                                "short": False
                            },
                            {
                                "title": "Operation",
                                "value": str(action).upper(),
                                "short": False
                            }
                        ]
                    }

                ]
            }
            if error is not None:
                payload['attachments'][0]['fields'].append({
                    "title": "Error",
                    "value": error,
                    "short": False
                })
            try:
                requests.post(slack_webhook_url, json=payload)
            except Exception as e:
                print(e)
                return False
    return True



import argparse
from argparse import RawTextHelpFormatter
import sys
import os
import yaml


PACKAGE_PARENT = '..'
SCRIPT_DIR = os.path.dirname(os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__))))
sys.path.append(os.path.normpath(os.path.join(SCRIPT_DIR, PACKAGE_PARENT)))

from stackuchin.create import create
from stackuchin.delete import delete
from stackuchin.update import update
from stackuchin.start_pipeline import start_pipeline


class StackuchinCLI(object):

    def __init__(self):
        parser = argparse.ArgumentParser(
            description='CLI tool to automatically create, update and delete AWS CloudFormation '
                        'stacks in multiple AWS accounts and regions at the same time',
            usage='''stackuchin <command> [<args>]

To see help text, you can run:
    stackuchin help
    stackuchin version
    stackuchin create --help
    stackuchin delete --help
    stackuchin update --help
    stackuchin pipeline --help
''')
        parser.add_argument('command', help='Command to run')
        # parse_args defaults to [1:] for args, but you need to
        # exclude the rest of the args too, or validation will fail
        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            parser.print_help()
            exit(1)
        # use dispatch pattern to invoke method with same name
        getattr(self, args.command)()

    # noinspection PyMethodMayBeStatic
    def version(self):
        print("1.5.4")

    # noinspection PyMethodMayBeStatic
    def create(self):
        parser = argparse.ArgumentParser(
            description='Create command system',
            formatter_class=RawTextHelpFormatter
        )
        parser.add_argument('--stack_file', dest="stack_file"
                            , default='./cloudformation-stacks.yaml'
                            , help="The YAML file which contains your stack definitions.\n"
                                   "Defaults to \"./cloudformation-stacks.yaml\" if not specified.")
        parser.add_argument('--stack_name', dest="stack_name", required=True
                            , help="The stack that you wish to create")
        parser.add_argument('--secret', dest="secret", required=False, default=None
                            , action='append', metavar='Parameter=Value'
                            , help='Argument used to specify values for NoEcho parameters in your stack')
        parser.add_argument('--slack_webhook', dest="slack_webhook", required=False, default=None
                            , help='Argument used to overwrite environment variable STACKUCHIN_SLACK.\n'
                                   'If argument is specified, any notifications will be sent to this URL.\n'
                                   'If not specified, the script will check for env var STACKUCHIN_SLACK.\n'
                                   'If neither argument nor environment variable is specified, then no notifications '
                                   'will be sent.')
        parser.add_argument('--s3_bucket', dest="s3_bucket", required=False, default=None
                            , help='Argument used to overwrite environment variable STACKUCHIN_BUCKET_NAME.\n'
                                   'If argument is specified, then the template is first uploaded here before '
                                   'used in the stack.\n'
                                   'If not specified, the script will check for env var STACKUCHIN_BUCKET_NAME.\n'
                                   'If neither argument nor environment variable is specified, then the script will '
                                   'attempt to feed the template directly to the AWS API call, however, due to '
                                   'AWS CloudFormation API call limitations, you might end up with a bigger template '
                                   'in byte size than the max value allowed by AWS.\n'
                                   'Details here -> https://docs.aws.amazon.com/AWSCloudFormation/latest/'
                                   'UserGuide/cloudformation-limits.html')
        parser.add_argument('--s3_prefix', dest="s3_prefix", required=False, default=None
                            , help='Argument used to overwrite environment variable STACKUCHIN_BUCKET_PREFIX.\n'
                                   'The bucket prefix path to be used when the S3 bucket is defined.')
        parser.add_argument('--only_errors', dest="only_errors", required=False, default=False, action="store_true"
                            , help='By default, all notifications are sent to Slack if slack_webhook is defined.\n'
                                   'By running this command you ensure that only errors are getting pushed.\n'
                                   'This is useful in case you don\'t want to see COMPLETE and START notifications.')
        parser.add_argument('-p', '--profile', dest='profile', default=None
                            , help='The AWS profile you\'ll be using.\n'
                                   'If not specified, the "default" profile will be used. \n'
                                   'If no profiles are defined, then the default AWS credential mechanism starts.\n')
        args = parser.parse_args(sys.argv[2:])

        slack_webhook_url = None
        if args.slack_webhook is not None:
            slack_webhook_url = args.slack_webhook
        else:
            if "STACKUCHIN_SLACK" in os.environ:
                slack_webhook_url = os.environ.get('STACKUCHIN_SLACK')

        s3_bucket = None
        if args.s3_bucket is not None:
            s3_bucket = args.s3_bucket
        else:
            if "STACKUCHIN_BUCKET_NAME" in os.environ:
                s3_bucket = os.environ.get('STACKUCHIN_BUCKET_NAME')

        s3_prefix = None
        if args.s3_prefix is not None:
            s3_prefix = args.s3_prefix
        else:
            if "STACKUCHIN_BUCKET_PREFIX" in os.environ:
                s3_prefix = os.environ.get('STACKUCHIN_BUCKET_PREFIX')

        create(args.profile, args.stack_file, args.stack_name,
               args.secret, slack_webhook_url, s3_bucket, s3_prefix, args.only_errors)

    # noinspection PyMethodMayBeStatic
    def delete(self):
        parser = argparse.ArgumentParser(
            description='Delete command system',
            formatter_class=RawTextHelpFormatter
        )
        parser.add_argument('--stack_file', dest="stack_file"
                            , default='./cloudformation-stacks.yaml'
                            , help="The YAML file which contains your stack definitions.\n"
                                   "Defaults to \"./cloudformation-stacks.yaml\" if not specified.")
        parser.add_argument('--stack_name', dest="stack_name", required=True
                            , help="The stack that you wish to create")
        parser.add_argument('--slack_webhook', dest="slack_webhook", required=False, default=None
                            , help='Argument used to overwrite environment variable STACKUCHIN_SLACK.\n'
                                   'If argument is specified, any notifications will be sent to this URL.\n'
                                   'If not specified, the script will check for env var STACKUCHIN_SLACK.\n'
                                   'If neither argument nor environment variable is specified, then no notifications '
                                   'will be sent.')
        parser.add_argument('--only_errors', dest="only_errors", required=False, default=False, action="store_true"
                            , help='By default, all notifications are sent to Slack if slack_webhook is defined.\n'
                                   'By running this command you ensure that only errors are getting pushed.\n'
                                   'This is useful in case you don\'t want to see COMPLETE and START notifications.')
        parser.add_argument('-p', '--profile', dest='profile', default=None
                            , help='The AWS profile you\'ll be using.\n'
                                   'If not specified, the "default" profile will be used. \n'
                                   'If no profiles are defined, then the default AWS credential mechanism starts.\n')
        args = parser.parse_args(sys.argv[2:])

        slack_webhook_url = None
        if args.slack_webhook is not None:
            slack_webhook_url = args.slack_webhook
        else:
            if "STACKUCHIN_SLACK" in os.environ:
                slack_webhook_url = os.environ.get('STACKUCHIN_SLACK')

        delete(args.profile, args.stack_file, args.stack_name, slack_webhook_url, args.only_errors)

    # noinspection PyMethodMayBeStatic
    def update(self):
        parser = argparse.ArgumentParser(
            description='Update command system',
            formatter_class=RawTextHelpFormatter
        )
        parser.add_argument('--stack_file', dest="stack_file"
                            , default='./cloudformation-stacks.yaml'
                            , help="The YAML file which contains your stack definitions.\n"
                                   "Defaults to \"./cloudformation-stacks.yaml\" if not specified.")
        parser.add_argument('--stack_name', dest="stack_name", required=True
                            , help="The stack that you wish to update")
        parser.add_argument('--secret', dest="secret", required=False, default=None
                            , action='append', metavar='Parameter=Value'
                            , help='Argument used to specify values for NoEcho parameters in your stack')
        parser.add_argument('--slack_webhook', dest="slack_webhook", required=False, default=None
                            , help='Argument used to overwrite environment variable STACKUCHIN_SLACK.\n'
                                   'If argument is specified, any notifications will be sent to this URL.\n'
                                   'If not specified, the script will check for env var STACKUCHIN_SLACK.\n'
                                   'If neither argument nor environment variable is specified, then no notifications '
                                   'will be sent.')
        parser.add_argument('--s3_bucket', dest="s3_bucket", required=False, default=None
                            , help='Argument used to overwrite environment variable STACKUCHIN_BUCKET_NAME.\n'
                                   'If argument is specified, then the template is first uploaded here before '
                                   'used in the stack.\n'
                                   'If not specified, the script will check for env var STACKUCHIN_BUCKET_NAME.\n'
                                   'If neither argument nor environment variable is specified, then the script will '
                                   'attempt to feed the template directly to the AWS API call, however, due to '
                                   'AWS CloudFormation API call limitations, you might end up with a bigger template '
                                   'in byte size than the max value allowed by AWS.\n'
                                   'Details here -> https://docs.aws.amazon.com/AWSCloudFormation/latest/'
                                   'UserGuide/cloudformation-limits.html')
        parser.add_argument('--s3_prefix', dest="s3_prefix", required=False, default=None
                            , help='Argument used to overwrite environment variable STACKUCHIN_BUCKET_PREFIX.\n'
                                   'The bucket prefix path to be used when the S3 bucket is defined.')
        parser.add_argument('--only_errors', dest="only_errors", required=False, default=False, action="store_true"
                            , help='By default, all notifications are sent to Slack if slack_webhook is defined.\n'
                                   'By running this command you ensure that only errors are getting pushed.\n'
                                   'This is useful in case you don\'t want to see COMPLETE and START notifications.')
        parser.add_argument('-p', '--profile', dest='profile', default=None
                            , help='The AWS profile you\'ll be using.\n'
                                   'If not specified, the "default" profile will be used. \n'
                                   'If no profiles are defined, then the default AWS credential mechanism starts.\n')
        args = parser.parse_args(sys.argv[2:])


        slack_webhook_url = None
        if args.slack_webhook is not None:
            slack_webhook_url = args.slack_webhook
        else:
            if "STACKUCHIN_SLACK" in os.environ:
                slack_webhook_url = os.environ.get('STACKUCHIN_SLACK')

        s3_bucket = None
        if args.s3_bucket is not None:
            s3_bucket = args.s3_bucket
        else:
            if "STACKUCHIN_BUCKET_NAME" in os.environ:
                s3_bucket = os.environ.get('STACKUCHIN_BUCKET_NAME')

        s3_prefix = None
        if args.s3_prefix is not None:
            s3_prefix = args.s3_prefix
        else:
            if "STACKUCHIN_BUCKET_PREFIX" in os.environ:
                s3_prefix = os.environ.get('STACKUCHIN_BUCKET_PREFIX')

        update(args.profile, args.stack_file, args.stack_name, args.secret,
               slack_webhook_url, s3_bucket, s3_prefix, args.only_errors)

    # noinspection PyMethodMayBeStatic
    def pipeline(self):
        parser = argparse.ArgumentParser(
            description='Create command system',
            formatter_class=RawTextHelpFormatter
        )
        parser.add_argument('--stack_file', dest="stack_file"
                            , default='./cloudformation-stacks.yaml'
                            , help="The YAML file which contains your stack definitions.\n"
                                   "Defaults to \"./cloudformation-stacks.yaml\" if not specified.")
        parser.add_argument('--pipeline_file', dest="pipeline_file", required=True
                            , help="The pipeline definition file to run your deployments.")
        parser.add_argument('--slack_webhook', dest="slack_webhook", required=False, default=None
                            , help='Argument used to overwrite environment variable STACKUCHIN_SLACK.\n'
                                   'If argument is specified, any notifications will be sent to this URL.\n'
                                   'If not specified, the script will check for env var STACKUCHIN_SLACK.\n'
                                   'If neither argument nor environment variable is specified, then no notifications '
                                   'will be sent.')
        parser.add_argument('--s3_bucket', dest="s3_bucket", required=False, default=None
                            , help='Argument used to overwrite environment variable STACKUCHIN_BUCKET_NAME.\n'
                                   'If argument is specified, then the template is first uploaded here before '
                                   'used in the stack.\n'
                                   'If not specified, the script will check for env var STACKUCHIN_BUCKET_NAME.\n'
                                   'If neither argument nor environment variable is specified, then the script will '
                                   'attempt to feed the template directly to the AWS API call, however, due to '
                                   'AWS CloudFormation API call limitations, you might end up with a bigger template '
                                   'in byte size than the max value allowed by AWS.\n'
                                   'Details here -> https://docs.aws.amazon.com/AWSCloudFormation/latest/'
                                   'UserGuide/cloudformation-limits.html')
        parser.add_argument('--s3_prefix', dest="s3_prefix", required=False, default=None
                            , help='Argument used to overwrite environment variable STACKUCHIN_BUCKET_PREFIX.\n'
                                   'The bucket prefix path to be used when the S3 bucket is defined.')
        parser.add_argument('--only_errors', dest="only_errors", required=False, default=False, action="store_true"
                            , help='By default, all notifications are sent to Slack if slack_webhook is defined.\n'
                                   'By running this command you ensure that only errors are getting pushed.\n'
                                   'This is useful in case you don\'t want to see COMPLETE and START notifications.')
        parser.add_argument('-p', '--profile', dest='profile', default=None
                            , help='The AWS profile you\'ll be using.\n'
                                   'If not specified, the "default" profile will be used. \n'
                                   'If no profiles are defined, then the default AWS credential mechanism starts.\n')
        args = parser.parse_args(sys.argv[2:])

        stacks = None
        try:
            with open(args.stack_file, 'r') as stack_stream:
                stacks = yaml.safe_load(stack_stream)
        except yaml.YAMLError as exc:
            print(exc)
            exit(1)

        pipeline = None
        try:
            with open(args.pipeline_file, 'r') as pipeline_stream:
                pipeline = yaml.safe_load(pipeline_stream)
        except yaml.YAMLError as exc:
            print(exc)
            exit(1)

        if 'pipeline' not in pipeline:
            print("The pipeline_file {} must contain a top-level object called \"pipeline\".".format(
                args.pipeline_file))
            exit(1)

        if 'pipeline_type' in pipeline['pipeline']:
            if str(pipeline['pipeline']['pipeline_type']).lower() not in ['parallel', 'sequential']:
                print("The value for \"pipeline_type\" can be either \"parallel\" or \"sequential\".")
                print("If not specified, the default value is \"sequential\".")
                exit(1)

        if 'update' not in pipeline['pipeline'] and \
                'delete' not in pipeline['pipeline'] and \
                'create' not in pipeline['pipeline']:
            print("An action type of either \"update\", \"create\", or \"delete\" must be defined "
                  "in the \"pipeline\" object definition.")
            exit(1)

        for action in ["update", "create", "delete"]:
            if action in pipeline["pipeline"]:
                if type(pipeline["pipeline"][action]) is not list:
                    print("Expected a list of inputs for the command {}.".format(action))
                    exit(1)
                for item in pipeline["pipeline"][action]:
                    if "stack_name" not in item:
                        print("A property with key \"stack_name\" must be present in each item "
                              "for the {} command.".format(action))
                        exit(1)
                    if "no_echo" in item:
                        if type(item["no_echo"]) is not list:
                            print("If you want to specify \"secrets\", make sure they "
                                  "are a list of Name/Value objects.")
                            exit(1)
                            for secret in item["no_echo"]:
                                if "Name" not in secret or "Value" not in secret:
                                    print("You must specify a combination of Name/Value objects "
                                          "for each item in your secrets")
                                    exit(1)

        slack_webhook_url = None
        if args.slack_webhook is not None:
            slack_webhook_url = args.slack_webhook
        else:
            if "STACKUCHIN_SLACK" in os.environ:
                slack_webhook_url = os.environ.get('STACKUCHIN_SLACK')

        s3_bucket = None
        if args.s3_bucket is not None:
            s3_bucket = args.s3_bucket
        else:
            if "STACKUCHIN_BUCKET_NAME" in os.environ:
                s3_bucket = os.environ.get('STACKUCHIN_BUCKET_NAME')

        s3_prefix = None
        if args.s3_prefix is not None:
            s3_prefix = args.s3_prefix
        else:
            if "STACKUCHIN_BUCKET_PREFIX" in os.environ:
                s3_prefix = os.environ.get('STACKUCHIN_BUCKET_PREFIX')

        start_pipeline(args.profile, args.stack_file, args.pipeline_file,
                       slack_webhook_url, s3_bucket, s3_prefix, args.only_errors)

def main():
    StackuchinCLI()


if __name__ == '__main__':
    StackuchinCLI()



import boto3
from botocore.exceptions import WaiterError
import yaml
from datetime import datetime
import simplejson as json

from stackuchin.utilities import alert
from stackuchin.utilities import result

import warnings
warnings.filterwarnings("ignore")


def update(profile_name, stack_file, stack_name, secret, slack_webhook_url,
           s3_bucket, s3_prefix, only_errors, from_pipeline=False):
    aws_region = None
    aws_account = None
    stacks = None
    stack_template = None
    action = 'UPDATE'
    changeset = None
    client_token = datetime.utcnow().isoformat().replace(":", "-").replace(".", "-")

    stacks = None
    try:
        with open(stack_file, 'r') as stack_stream:
            stacks = yaml.safe_load(stack_stream)
    except yaml.YAMLError as exc:
        print(exc)
        exit(1)

    if stack_name not in stacks:
        print("{} was not found in your stack file.".format(stack_name))
        if from_pipeline:
            alert(stack_name,
                  "{} was not found in your stack file {}.".format(
                      stack_name, stack_file),
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    if 'Account' not in stacks[stack_name]:
        print("The Account property is missing from the stack definition.")
        if from_pipeline:
            alert(stack_name,
                  "The Account property is missing from the stack {} in stack_file {}.".format(
                      stack_name, stack_file),
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    if 'Region' not in stacks[stack_name]:
        print("The Region property is missing from the stack definition.")
        if from_pipeline:
            alert(stack_name,
                  "The Region property is missing from the stack {} in stack_file {}.".format(
                      stack_name, stack_file),
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    if 'Template' not in stacks[stack_name]:
        print("The Template property is missing from the stack definition.")
        if from_pipeline:
            alert(stack_name,
                  "The Template property is missing from the stack {} in stack_file {}.".format(
                      stack_name, stack_file),
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    if 'Parameters' not in stacks[stack_name]:
        print("The Parameters property is missing from the stack definition.")
        print("Should you with to deploy a stack with no parameters, please define the property as: \n"
              "\"Parameters: {}\"")
        if from_pipeline:
            alert(stack_name,
                  "The Parameters property is missing from the stack "
                  "" + stack_name + " in stack_file " + stack_file + ".\n"
                  "Should you with to deploy a stack with no parameters, please simply define the property as: \n"
                  "Parameters: {}",
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    template = None
    try:
        with open(stacks[stack_name]['Template'], 'r') as template_stream:
            template = json.load(template_stream)
    except Exception as e:
        try:
            with open(stacks[stack_name]['Template'], 'r') as template_stream:
                template = yaml.safe_load(template_stream)
        except yaml.YAMLError as exc:
            print(exc)
            exit(1)

    if 'Resources' not in template:
        print("The CloudFormation templates provided for this stack does not contain any Resources")
        if from_pipeline:
            alert(stack_name,
                  "The CloudFormation template {} for stack {} does not contain any Resources.".format(
                      stacks[stack_name]['Template'], stack_name),
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    # Get auth
    if not from_pipeline:
        if profile_name is not None:
            try:
                boto3.setup_default_session(profile_name=profile_name)
            except Exception as exc:
                print(exc)
                exit(1)

    # Verify integrity of stack
    stacks = None
    try:
        with open(stack_file, 'r') as stack_stream:
            stacks = yaml.safe_load(stack_stream)
            aws_region = stacks[stack_name]['Region']
            aws_account = stacks[stack_name]['Account']
            stack_template = stacks[stack_name]['Template']
    except yaml.YAMLError as exc:
        print(exc)
        alert(stack_name,
              "Unable to verify required options for stack {} in stack file {}. Exception = {}".format(
                  stack_name, stack_file, exc),
              aws_region, aws_account, action, profile_name, slack_webhook_url)
        exit(1)

    # Connect to AWS CloudFormation
    cf_client = boto3.client('cloudformation', region_name=aws_region)

    # Get parameter and tags from stack
    stack_parameters, stack_tags, stack_template_url = result(stack_file, stack_name, secret,
                                                              s3_bucket, s3_prefix,
                                                              aws_region, aws_account,
                                                              action, profile_name, slack_webhook_url)

    # Create change set
    try:
        if stack_template_url['type'] == 'TemplateURL':
            changeset = cf_client.create_change_set(StackName=stack_name,
                                                    TemplateURL=stack_template_url['value'],
                                                    Parameters=stack_parameters,
                                                    Capabilities=[
                                                        'CAPABILITY_NAMED_IAM', 'CAPABILITY_IAM',
                                                        'CAPABILITY_AUTO_EXPAND'
                                                    ],
                                                    Tags=stack_tags,
                                                    ClientToken=client_token,
                                                    ChangeSetName="Changeset-{}".format(client_token))
        else:
            template_body = stack_template_url['value']
            if type(template_body) != str:
                template_body = json.dumps(template_body)
            changeset = cf_client.create_change_set(StackName=stack_name,
                                                    TemplateBody=template_body,
                                                    Parameters=stack_parameters,
                                                    Capabilities=[
                                                        'CAPABILITY_NAMED_IAM', 'CAPABILITY_IAM',
                                                        'CAPABILITY_AUTO_EXPAND'
                                                    ],
                                                    Tags=stack_tags,
                                                    ClientToken=client_token,
                                                    ChangeSetName="Changeset-{}".format(client_token))
    except Exception as exc:
        print(exc)
        alert(stack_name,
              "Unable to create change set. Exception = {}".format(exc),
              aws_region, aws_account, action, profile_name, slack_webhook_url)
        exit(1)

    # Waiting for the change set to finish creating
    waiter = None
    try:
        waiter = cf_client.get_waiter('change_set_create_complete')
        waiter.wait(ChangeSetName=changeset['Id'])
    except WaiterError as exc:
        if 'Submit different information' in exc.last_response['StatusReason']:
            print("Nothing to update for stack {}".format(stack_name))
            if from_pipeline:
                return True
            exit(0)
        else:
            print(exc)
            alert(stack_name,
                  "Unable to create change set. Exception = {}".format(exc),
                  aws_region, aws_account, action, profile_name, slack_webhook_url)
            exit(1)

    # Executing change set
    try:
        cf_client.execute_change_set(ChangeSetName=changeset['Id'], ClientRequestToken=client_token)
    except Exception as exc:
        print(exc)
        alert(stack_name,
              "Unable to start stack creation process. Exception = {}".format(exc),
              aws_region, aws_account, action, profile_name, slack_webhook_url)
        exit(1)

    # Waiting for stack to finish updating
    if not only_errors:
        print("UPDATE_STARTED for stack {}".format(stack_name))
        alert(stack_name, None, aws_region, aws_account, "UPDATE_STARTED", profile_name, slack_webhook_url)
    waiter = None
    try:
        waiter = cf_client.get_waiter('stack_update_complete')
        waiter.wait(StackName=stack_name)
    except Exception as exc:
        pass

    # Check final status of stack
    stack_events = cf_client.describe_stack_events(StackName=stack_name)
    failure_reasons = []
    for resource in stack_events["StackEvents"]:
        if "ClientRequestToken" in resource and \
            client_token == resource["ClientRequestToken"] and \
                "FAILED" in str(resource["ResourceStatus"]):

            failure_reasons.append("- *{}* <> {}".format(resource["LogicalResourceId"],
                                                         resource["ResourceStatusReason"]))
    if len(failure_reasons) > 0:
        print("UPDATE_FAILURE for stack {}".format(stack_name))
        alert(stack_name,
              'Stack creation process ended with status {}.\n'
              'Failure reasons :\n{}'.format(
                  "UPDATE_FAILED",
                  "\n".join(failure_reasons)
              ),
              aws_region, aws_account, action, profile_name, slack_webhook_url)
        exit(1)

    if not only_errors:
        print("UPDATE_COMPLETE for stack {}".format(stack_name))
        alert(stack_name, None, aws_region, aws_account, "UPDATE_COMPLETE", profile_name, slack_webhook_url)

    return True



import boto3
import yaml
from datetime import datetime
import simplejson as json

from stackuchin.utilities import alert

import warnings

warnings.filterwarnings("ignore")


def delete(profile_name, stack_file, stack_name,
           slack_webhook_url, only_errors, from_pipeline=False):
    aws_region = None
    aws_account = None
    stacks = None
    stack_template = None
    action = 'DELETE'
    client_token = datetime.utcnow().isoformat().replace(":", "-").replace(".", "-")

    stacks = None
    try:
        with open(stack_file, 'r') as stack_stream:
            stacks = yaml.safe_load(stack_stream)
    except yaml.YAMLError as exc:
        print(exc)
        exit(1)

    if stack_name not in stacks:
        print("{} was not found in your stack file.".format(stack_name))
        if from_pipeline:
            alert(stack_name,
                  "{} was not found in your stack file {}.".format(
                      stack_name, stack_file),
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    if 'Account' not in stacks[stack_name]:
        print("The Account property is missing from the stack definition.")
        if from_pipeline:
            alert(stack_name,
                  "The Account property is missing from the stack {} in stack_file {}.".format(
                      stack_name, stack_file),
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    if 'Region' not in stacks[stack_name]:
        print("The Region property is missing from the stack definition.")
        if from_pipeline:
            alert(stack_name,
                  "The Region property is missing from the stack {} in stack_file {}.".format(
                      stack_name, stack_file),
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    if 'Template' not in stacks[stack_name]:
        print("The Template property is missing from the stack definition.")
        if from_pipeline:
            alert(stack_name,
                  "The Template property is missing from the stack {} in stack_file {}.".format(
                      stack_name, stack_file),
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    if 'Parameters' not in stacks[stack_name]:
        print("The Parameters property is missing from the stack definition.")
        print("Should you with to deploy a stack with no parameters, please define the property as: \n"
              "\"Parameters: {}\"")
        if from_pipeline:
            alert(stack_name,
                  "The Parameters property is missing from the stack "
                  "" + stack_name + " in stack_file " + stack_file + ".\n"
                  "Should you with to deploy a stack with no parameters, please simply define the property as: \n"
                  "Parameters: {}",
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    template = None
    try:
        with open(stacks[stack_name]['Template'], 'r') as template_stream:
            template = json.load(template_stream)
    except Exception as e:
        try:
            with open(stacks[stack_name]['Template'], 'r') as template_stream:
                template = yaml.safe_load(template_stream)
        except yaml.YAMLError as exc:
            print(exc)
            exit(1)

    if 'Resources' not in template:
        print("The CloudFormation templates provided for this stack does not contain any Resources")
        if from_pipeline:
            alert(stack_name,
                  "The CloudFormation template {} for stack {} does not contain any Resources.".format(
                      stacks[stack_name]['Template'], stack_name),
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    # Get auth
    if not from_pipeline:
        if profile_name is not None:
            try:
                boto3.setup_default_session(profile_name=profile_name)
            except Exception as exc:
                print(exc)
                exit(1)

    # Verify integrity of stack
    stacks = None
    try:
        with open(stack_file, 'r') as stack_stream:
            stacks = yaml.safe_load(stack_stream)
            aws_region = stacks[stack_name]['Region']
            aws_account = stacks[stack_name]['Account']
            stack_template = stacks[stack_name]['Template']
    except yaml.YAMLError as exc:
        print(exc)
        alert(stack_name,
              "Unable to verify required options for stack {} in stack file {}. Exception = {}".format(
                  stack_name, stack_file, exc),
              aws_region, aws_account, action, profile_name, slack_webhook_url)
        exit(1)

    # Connect to AWS CloudFormation
    cf_client = boto3.client('cloudformation', region_name=aws_region)

    # Check if stack exists
    stack_id = None
    try:
        stack_check = cf_client.describe_stacks(
            StackName=stack_name
        )
        stack_id = stack_check['Stacks'][0]['StackId']
    except Exception as exc:
        print(exc)
        alert(stack_name,
              "Unable to check if stack already exists. Exception = {}".format(exc),
              aws_region, aws_account, action, profile_name, slack_webhook_url)
        exit(1)

    # Disable termination protection
    try:
        cf_client.update_termination_protection(StackName=stack_name, EnableTerminationProtection=False)
    except Exception as exc:
        print(exc)
        alert(stack_name,
              "Unable to disable termination protection. Exception = {}".format(exc),
              aws_region, aws_account, action, profile_name, slack_webhook_url)
        exit(1)

    # Delete the stack
    waiter = None
    if not only_errors:
        print("DELETE_STARTED for stack {}".format(stack_name))
        alert(stack_name, None, aws_region, aws_account, "DELETE_STARTED", profile_name, slack_webhook_url)
    try:
        cf_client.delete_stack(StackName=stack_name, ClientRequestToken=client_token)
    except Exception as exc:
        print(exc)
        alert(stack_name,
              "Unable to start stack deletion process. Exception = {}".format(exc),
              aws_region, aws_account, action, profile_name, slack_webhook_url)
        exit(1)

    # Waiting for stack to finish deleting
    try:
        waiter = cf_client.get_waiter('stack_delete_complete')
        waiter.wait(StackName=stack_name)
    except Exception as exc:
        pass

    # Check final status of stack
    stack_events = cf_client.describe_stack_events(StackName=stack_id)
    failure_reasons = []
    for resource in stack_events["StackEvents"]:
        if "ClientRequestToken" in resource and \
            client_token == resource["ClientRequestToken"] and \
                "FAILED" in str(resource["ResourceStatus"]):

            failure_reasons.append("- *{}* <> {}".format(resource["LogicalResourceId"],
                                                         resource["ResourceStatusReason"]))
    if len(failure_reasons) > 0:
        print("DELETE_FAILURE for stack {}".format(stack_name))
        alert(stack_name,
              'Stack deletion process ended with status {}.\n'
              'Failure reasons :\n{}'.format(
                  "DELETE_FAILED",
                  "\n".join(failure_reasons)
              ),
              aws_region, aws_account, action, profile_name, slack_webhook_url)
        exit(1)

    if not only_errors:
        print("DELETE_COMPLETE for stack {}".format(stack_name))
        alert(stack_name, None, aws_region, aws_account, "DELETE_COMPLETE", profile_name, slack_webhook_url)

    return True



import yaml
import concurrent.futures
import boto3
import time
import random


from stackuchin.create import create
from stackuchin.delete import delete
from stackuchin.update import update

import warnings
warnings.filterwarnings("ignore")


def perform_parallel_changes(stack):
    time.sleep(random.randint(0, 30))

    if stack['action'] == 'create':
        return create(
            stack['profile_name'],
            stack['stack_file'],
            stack["stack_name"],
            stack['secrets'],
            stack['slack_webhook_url'],
            stack['s3_bucket'],
            stack['s3_prefix'],
            stack['only_errors'],
            True)

    if stack['action'] == 'update':
        return update(
            stack['profile_name'],
            stack['stack_file'],
            stack["stack_name"],
            stack['secrets'],
            stack['slack_webhook_url'],
            stack['s3_bucket'],
            stack['s3_prefix'],
            stack['only_errors'],
            True)

    if stack['action'] == 'delete':
        return delete(
            stack['profile_name'],
            stack['stack_file'],
            stack["stack_name"],
            stack['slack_webhook_url'],
            stack['only_errors'],
            True)

    return True


def start_pipeline(profile_name, stack_file, pipeline_file,
                   slack_webhook_url, s3_bucket, s3_prefix, only_errors):

    pipeline_type = 'sequential'

    pipeline = None
    try:
        with open(pipeline_file, 'r') as pipeline_stream:
            pipeline = yaml.safe_load(pipeline_stream)
            pipeline = pipeline['pipeline']
    except yaml.YAMLError as exc:
        print(exc)
        exit(1)

    if 'pipeline_type' in pipeline:
        pipeline_type = pipeline['pipeline_type']

    if profile_name is not None:
        boto3.setup_default_session(profile_name=profile_name)

    if pipeline_type == "sequential":
        # First create new stacks (if any)
        if "create" in pipeline:
            for stack in pipeline["create"]:
                secrets = []
                if "secrets" in stack:
                    for hidden in stack["secrets"]:
                        secrets.append(
                            "{}={}".format(hidden["Name"], hidden["Value"])
                        )
                create(profile_name, stack_file, stack["stack_name"], secrets,
                       slack_webhook_url, s3_bucket, s3_prefix, only_errors, True)

        # Then update existing stacks (if any)
        if "update" in pipeline:
            for stack in pipeline["update"]:
                secrets = []
                if "secrets" in stack:
                    for hidden in stack["secrets"]:
                        secrets.append(
                            "{}={}".format(hidden["Name"], hidden["Value"])
                        )
                update(profile_name, stack_file, stack["stack_name"], secrets,
                       slack_webhook_url, s3_bucket, s3_prefix, only_errors, True)

        # And finally delete stacks (if any)
        if "delete" in pipeline:
            for stack in pipeline["delete"]:
                delete(profile_name, stack_file, stack["stack_name"], slack_webhook_url, only_errors, True)

    else:
        # Build array of executions
        stacks = []

        # Check if any stacks to create
        if "create" in pipeline:
            for stack in pipeline["create"]:
                secrets = []
                if "secrets" in stack:
                    for hidden in stack["secrets"]:
                        secrets.append(
                            "{}={}".format(hidden["Name"], hidden["Value"])
                        )
                stacks.append({
                    "action": "create",
                    "profile_name": profile_name,
                    "stack_file": stack_file,
                    "stack_name": stack["stack_name"],
                    "secrets": secrets,
                    "slack_webhook_url": slack_webhook_url,
                    "s3_bucket": s3_bucket,
                    "s3_prefix": s3_prefix,
                    "only_errors": only_errors
                })

        # Check if any stacks to update
        if "update" in pipeline:
            for stack in pipeline["update"]:
                secrets = []
                if "secrets" in stack:
                    for hidden in stack["secrets"]:
                        secrets.append(
                            "{}={}".format(hidden["Name"], hidden["Value"])
                        )
                stacks.append({
                    "action": "update",
                    "profile_name": profile_name,
                    "stack_file": stack_file,
                    "stack_name": stack["stack_name"],
                    "secrets": secrets,
                    "slack_webhook_url": slack_webhook_url,
                    "s3_bucket": s3_bucket,
                    "s3_prefix": s3_prefix,
                    "only_errors": only_errors
                })

        # Check if any stacks to delete
        if "delete" in pipeline:
            for stack in pipeline["delete"]:
                secrets = []
                if "secrets" in stack:
                    for hidden in stack["secrets"]:
                        secrets.append(
                            "{}={}".format(hidden["Name"], hidden["Value"])
                        )
                stacks.append({
                    "action": "delete",
                    "profile_name": profile_name,
                    "stack_file": stack_file,
                    "stack_name": stack["stack_name"],
                    "secrets": secrets,
                    "slack_webhook_url": slack_webhook_url,
                    "s3_bucket": s3_bucket,
                    "s3_prefix": s3_prefix,
                    "only_errors": only_errors
                })

        # Execute parallel deployments
        with concurrent.futures.ThreadPoolExecutor(
                max_workers=len(stacks)) as executor:
            results = {
                executor.submit(perform_parallel_changes, stack): stack
                for stack in stacks
            }
            for future in concurrent.futures.as_completed(results):
                future.result()




# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

current_version = str('1.5.4')

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='stackuchin',

    version=current_version,

    description='CLI for Stackuchin - Automatically create, update and delete AWS CloudFormation stacks '
                'in multiple AWS accounts and regions at the same time',

    long_description=long_description,

    long_description_content_type='text/markdown',

    url='https://github.com/Rungutan/stackuchin',

    download_url='https://github.com/Rungutan/stackuchin/archive/{}.tar.gz'.format(current_version),

    author='Rungutan',

    author_email='support@rungutan.com',

    classifiers=[  # Optional
        # How mature is this project?
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Audience
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',

        'Topic :: Software Development :: Testing',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing :: Acceptance',
        'Topic :: Software Development :: Testing :: BDD',
        'Topic :: Software Development :: Testing :: Mocking',
        'Topic :: Software Development :: Testing :: Traffic Generation',
        'Topic :: Software Development :: Testing :: Unit',

        # License
        'License :: OSI Approved :: MIT License',

        # Supported Python versions
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
    ],

    keywords='stackuchin stackuchin-cli stackuchin_cli cli aws cloudformation stacks deployment integration'
             'cotinuous continuos devops tools',

    package_dir={'': 'src'},

    packages=find_packages(where='src'),

    python_requires='>=3.5, <4',

    install_requires=['simplejson', 'boto3', 'botocore', 'pyyaml', 'requests'],

    extras_require={
        'test': ['coverage'],
    },

    entry_points={
        'console_scripts': [
            'stackuchin=stackuchin:main',
        ],
    },

    project_urls={
        'Bug Reports': 'https://github.com/Rungutan/stackuchin/issues',
        'Source': 'https://github.com/Rungutan/stackuchin/',
    }
)



import boto3
import botocore
import yaml
from datetime import datetime
import simplejson as json

from stackuchin.utilities import alert
from stackuchin.utilities import result
from stackuchin.delete import delete

import warnings
warnings.filterwarnings("ignore")


def create(profile_name, stack_file, stack_name, secret, slack_webhook_url,
           s3_bucket, s3_prefix, only_errors, from_pipeline=False):
    aws_region = None
    aws_account = None
    stacks = None
    stack_template = None
    action = 'CREATE'
    client_token = datetime.utcnow().isoformat().replace(":", "-").replace(".", "-")

    stacks = None
    try:
        with open(stack_file, 'r') as stack_stream:
            stacks = yaml.safe_load(stack_stream)
    except yaml.YAMLError as exc:
        print(exc)
        exit(1)

    if stack_name not in stacks:
        print("{} was not found in your stack file.".format(stack_name))
        if from_pipeline:
            alert(stack_name,
                  "{} was not found in your stack file {}.".format(
                      stack_name, stack_file),
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    if 'Account' not in stacks[stack_name]:
        print("The Account property is missing from the stack definition.")
        if from_pipeline:
            alert(stack_name,
                  "The Account property is missing from the stack {} in stack_file {}.".format(
                      stack_name, stack_file),
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    if 'Region' not in stacks[stack_name]:
        print("The Region property is missing from the stack definition.")
        if from_pipeline:
            alert(stack_name,
                  "The Region property is missing from the stack {} in stack_file {}.".format(
                      stack_name, stack_file),
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    if 'Template' not in stacks[stack_name]:
        print("The Template property is missing from the stack definition.")
        if from_pipeline:
            alert(stack_name,
                  "The Template property is missing from the stack {} in stack_file {}.".format(
                      stack_name, stack_file),
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    if 'Parameters' not in stacks[stack_name]:
        print("The Parameters property is missing from the stack definition.")
        print("Should you with to deploy a stack with no parameters, please define the property as: \n"
              "\"Parameters: {}\"")
        if from_pipeline:
            alert(stack_name,
                  "The Parameters property is missing from the stack "
                  "" + stack_name + " in stack_file " + stack_file + ".\n "
                  "Should you with to deploy a stack with no parameters, please simply define the property as: \n"
                  "Parameters: {}",
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    template = None
    try:
        with open(stacks[stack_name]['Template'], 'r') as template_stream:
            template = json.load(template_stream)
    except Exception as e:
        try:
            with open(stacks[stack_name]['Template'], 'r') as template_stream:
                template = yaml.safe_load(template_stream)
        except yaml.YAMLError as exc:
            print(exc)
            exit(1)

    if 'Resources' not in template:
        print("The CloudFormation templates provided for this stack does not contain any Resources")
        if from_pipeline:
            alert(stack_name,
                  "The CloudFormation template {} for stack {} does not contain any Resources.".format(
                      stacks[stack_name]['Template'], stack_name),
                  None, None, action, profile_name, slack_webhook_url)
        exit(1)

    # Get auth
    if not from_pipeline:
        if profile_name is not None:
            try:
                boto3.setup_default_session(profile_name=profile_name)
            except Exception as exc:
                print(exc)
                exit(1)

    # Verify integrity of stack
    stacks = None
    try:
        with open(stack_file, 'r') as stack_stream:
            stacks = yaml.safe_load(stack_stream)
            aws_region = stacks[stack_name]['Region']
            aws_account = stacks[stack_name]['Account']
            stack_template = stacks[stack_name]['Template']
    except yaml.YAMLError as exc:
        print(exc)
        alert(stack_name,
              "Unable to verify required options for stack {} in stack file {}. Exception = {}".format(
                  stack_name, stack_file, exc),
              aws_region, aws_account, action, profile_name, slack_webhook_url)
        exit(1)

    # Connect to AWS CloudFormation
    cf_client = boto3.client('cloudformation', region_name=aws_region)

    # Get parameter and tags from stack
    stack_parameters, stack_tags, stack_template_url = result(stack_file, stack_name, secret,
                                                              s3_bucket, s3_prefix,
                                                              aws_region, aws_account,
                                                              action, profile_name, slack_webhook_url)

    # Check if stack exists and if it is in "ROLLBACK_COMPLETE" status, delete it
    try:
        stack_check = cf_client.describe_stacks(
            StackName=stack_name
        )
        if len(stack_check["Stacks"]) > 0:
            try:
                stack_check_status = stack_check["Stacks"][0]["StackStatus"]
                if stack_check_status == "ROLLBACK_COMPLETE":
                    # Delete the stack
                    delete(profile_name, stack_file, stack_name, slack_webhook_url)

            except Exception as exc:
                print(exc)
                alert(stack_name, exc, aws_region, aws_account, action, profile_name, slack_webhook_url)
                exit(1)
    except botocore.exceptions.ClientError as exc:
        if "Stack with id {} does not exist".format(stack_name) in str(exc):
            pass
        else:
            print(exc)
            alert(stack_name,
                  "Unable to check if stack already exists. Exception = {}".format(exc),
                  aws_region, aws_account, action, profile_name, slack_webhook_url)
            exit(1)
    except Exception as exc:
        print(exc)
        alert(stack_name,
              "Unable to check if stack already exists. Exception = {}".format(exc),
              aws_region, aws_account, action, profile_name, slack_webhook_url)
        exit(1)

    # Create the stack
    if not only_errors:
        print("CREATE_STARTED for stack {}".format(stack_name))
        alert(stack_name, None, aws_region, aws_account, "CREATE_STARTED", profile_name, slack_webhook_url)
    waiter = None
    try:
        if stack_template_url['type'] == 'TemplateURL':
            cf_client.create_stack(StackName=stack_name,
                                   TemplateURL=stack_template_url['value'],
                                   Parameters=stack_parameters,
                                   Capabilities=[
                                       'CAPABILITY_NAMED_IAM', 'CAPABILITY_IAM',
                                       'CAPABILITY_AUTO_EXPAND'
                                   ],
                                   Tags=stack_tags,
                                   EnableTerminationProtection=True,
                                   ClientRequestToken=client_token)
        else:
            cf_client.create_stack(StackName=stack_name,
                                   TemplateBody=stack_template_url['value'],
                                   Parameters=stack_parameters,
                                   Capabilities=[
                                       'CAPABILITY_NAMED_IAM', 'CAPABILITY_IAM',
                                       'CAPABILITY_AUTO_EXPAND'
                                   ],
                                   Tags=stack_tags,
                                   EnableTerminationProtection=True,
                                   ClientRequestToken=client_token)
    except Exception as exc:
        print(exc)
        alert(stack_name,
              "Unable to start stack creation process. Exception = {}".format(exc),
              aws_region, aws_account, action, profile_name, slack_webhook_url)
        exit(1)

    # Waiting for stack to finish creating
    try:
        waiter = cf_client.get_waiter('stack_create_complete')
        waiter.wait(StackName=stack_name)
    except Exception as exc:
        pass

    # Check final status of stack
    stack_events = cf_client.describe_stack_events(StackName=stack_name)
    failure_reasons = []
    for resource in stack_events["StackEvents"]:
        if "ClientRequestToken" in resource and \
            client_token == resource["ClientRequestToken"] and \
                "FAILED" in str(resource["ResourceStatus"]):

            failure_reasons.append("- *{}* <> {}".format(resource["LogicalResourceId"],
                                                         resource["ResourceStatusReason"]))
    if len(failure_reasons) > 0:
        print("CREATE_FAILURE for stack {}".format(stack_name))
        alert(stack_name,
              'Stack creation process ended with status {}.\n'
              'Failure reasons :\n{}'.format(
                  "CREATE_FAILED",
                  "\n".join(failure_reasons)
              ),
              aws_region, aws_account, action, profile_name, slack_webhook_url)
        exit(1)

    if not only_errors:
        print("CREATE_COMPLETE for stack {}".format(stack_name))
        alert(stack_name, None, aws_region, aws_account, "CREATE_COMPLETE", profile_name, slack_webhook_url)

    return True
